#ifndef RIME_PHASE_OP_H_
#define RIME_PHASE_OP_H_

#include "montblanc_macros.h"
namespace montblanc {
namespace phase {

// Declare the fully templated Phase class type up front
template <typename Device, typename FT, typename CT> class Phase;

} // namespace phase {
} // namespace montblanc {

#endif // #define RIME_PHASE_OP_H