#ifndef RIME_B_SQRT_OP_H_
#define RIME_B_SQRT_OP_H_

#include "montblanc_macros.h"

namespace montblanc {
namespace bsqrt {

template <typename Device, typename FT, typename CT> class BSqrt;

}
}

#endif